#!/usr/bin/python

import re as _re
import subprocess as _subprocess
import struct as _struct


def _get_stdout(args, stdin=None):
    stdin_pipe = None
    if stdin is not None:
        stdin_pipe = _subprocess.PIPE
    p = _subprocess.Popen(args, stdin=stdin_pipe, stdout=_subprocess.PIPE)
    stdout, stderr = p.communicate(stdin)
    status = p.wait()
    if status != 0:
        raise RuntimeError(status)
    return stdout


class PGPPacket (dict):
    # http://tools.ietf.org/search/rfc4880
    _old_format_packet_length_type = {  # type: (bytes, struct type)
        0: (1, 'B'),  # 1-byte unsigned integer
        1: (2, 'H'),  # 2-byte unsigned integer
        2: (4, 'I'),  # 4-byte unsigned integer
        3: (None, None),
        }

    _packet_types = {
        0: 'reserved',
        1: 'public-key encrypted session key packet',
        2: 'signature packet',
        3: 'symmetric-key encrypted session key packet',
        4: 'one-pass signature packet',
        5: 'secret-key packet',
        6: 'public-key packet',
        7: 'secret-subkey packet',
        8: 'compressed data packet',
        9: 'symmetrically encrypted data packet',
        10: 'marker packet',
        11: 'literal data packet',
        12: 'trust packet',
        13: 'user id packet',
        14: 'public-subkey packet',
        17: 'user attribute packet',
        18: 'sym. encrypted and integrity protected data packet',
        19: 'modification detection code packet',
        60: 'private',
        61: 'private',
        62: 'private',
        63: 'private',
        }

    _public_key_algorithms = {
        1: 'rsa (encrypt or sign)',
        2: 'rsa encrypt-only',
        3: 'rsa sign-only',
        16: 'elgamal (encrypt-only)',
        17: 'dsa (digital signature algorithm)',
        18: 'reserved for elliptic curve',
        19: 'reserved for ecdsa',
        20: 'reserved (formerly elgamal encrypt or sign)',
        21: 'reserved for diffie-hellman',
        100: 'private',
        101: 'private',
        102: 'private',
        103: 'private',
        104: 'private',
        105: 'private',
        106: 'private',
        107: 'private',
        108: 'private',
        109: 'private',
        110: 'private',
        }

    _symmetric_key_algorithms = {
        0: 'plaintext or unencrypted data',
        1: 'idea',
        2: 'tripledes',
        3: 'cast5',
        4: 'blowfish',
        5: 'reserved',
        6: 'reserved',
        7: 'aes with 128-bit key',
        8: 'aes with 192-bit key',
        9: 'aes with 256-bit key',
        10: 'twofish',
        100: 'private',
        101: 'private',
        102: 'private',
        103: 'private',
        104: 'private',
        105: 'private',
        106: 'private',
        107: 'private',
        108: 'private',
        109: 'private',
        110: 'private',
        }

    _compression_algorithms = {
        0: 'uncompressed',
        1: 'zip',
        2: 'zlib',
        3: 'bzip2',
        100: 'private',
        101: 'private',
        102: 'private',
        103: 'private',
        104: 'private',
        105: 'private',
        106: 'private',
        107: 'private',
        108: 'private',
        109: 'private',
        110: 'private',
        }

    _hash_algorithms = {
        1: 'md5',
        2: 'sha-1',
        3: 'ripe-md/160',
        4: 'reserved',
        5: 'reserved',
        6: 'reserved',
        7: 'reserved',
        8: 'sha256',
        9: 'sha384',
        10: 'sha512',
        11: 'sha224',
        100: 'private',
        101: 'private',
        102: 'private',
        103: 'private',
        104: 'private',
        105: 'private',
        106: 'private',
        107: 'private',
        108: 'private',
        109: 'private',
        110: 'private',
        }

    _string_to_key_types = {
        0: 'simple',
        1: 'salted',
        2: 'reserved',
        3: 'iterated and salted',
        100: 'private',
        101: 'private',
        102: 'private',
        103: 'private',
        104: 'private',
        105: 'private',
        106: 'private',
        107: 'private',
        108: 'private',
        109: 'private',
        110: 'private',
        }

    _clean_type_regex = _re.compile('\W+')

    def _clean_type(self):
        return self._clean_type_regex.sub('_', self['type'])

    def from_bytes(self, data):
        offset = self._parse_header(data=data)
        packet = data[offset:offset + self['length']]
        if len(packet) < self['length']:
            raise ValueError('packet too short ({} < {})'.format(
                len(packet), self['length']))
        offset += self['length']
        method_name = '_parse_{}'.format(self._clean_type())
        method = getattr(self, method_name, None)
        if not method:
            raise NotImplementedError(
                'cannot parse packet type {!r}'.format(self['type']))
        method(data=packet)
        return offset

    def _parse_header(self, data):
        packet_tag = data[0]
        offset = 1
        always_one = packet_tag & 1 << 7
        if not always_one:
            raise ValueError('most significant packet tag bit not set')
        self['new-format'] = packet_tag & 1 << 6
        if self['new-format']:
            type_code = packet_tag & 0b111111
            raise NotImplementedError('new-format packet length')
        else:
            type_code = packet_tag >> 2 & 0b1111
            self['length-type'] = packet_tag & 0b11
            length_bytes, length_type = self._old_format_packet_length_type[
                self['length-type']]
            if not length_bytes:
                raise NotImplementedError(
                    'old-format packet of indeterminate length')
            length_format = '>{}'.format(length_type)
            length_data = data[offset: offset + length_bytes]
            offset += length_bytes
            self['length'] = _struct.unpack(length_format, length_data)[0]
        self['type'] = self._packet_types[type_code]
        return offset

    @staticmethod
    def _parse_multiprecision_integer(data):
        r"""Parse RFC 4880's multiprecision integers

        >>> PGPPacket._parse_multiprecision_integer(b'\x00\x01\x01')
        (3, 1)
        >>> PGPPacket._parse_multiprecision_integer(b'\x00\x09\x01\xff')
        (4, 511)
        """
        bits = _struct.unpack('>H', data[:2])[0]
        offset = 2
        length = (bits + 7) // 8
        value = 0
        for i in range(length):
            value += data[offset + i] * 1 << (8 * (length - i - 1))
        offset += length
        return (offset, value)

    def _parse_string_to_key_specifier(self, data):
        self['string-to-key-type'] = self._string_to_key_types[data[0]]
        offset = 1
        if self['string-to-key-type'] == 'simple':
            self['string-to-key-hash-algorithm'] = self._hash_algorithms[
                data[offset]]
            offset += 1
        else:
            raise NotImplementedError(
                'string-to-key type {}'.format(self['string-to-key-type']))
        return offset

    def _parse_public_key_packet(self, data):
        self._parse_generic_public_key_packet(data=data)

    def _parse_public_subkey_packet(self, data):
        self._parse_generic_public_key_packet(data=data)

    def _parse_generic_public_key_packet(self, data):
        self['key-version'] = data[0]
        offset = 1
        if self['key-version'] != 4:
            raise NotImplementedError(
                'public (sub)key packet version {}'.format(
                    self['key-version']))
        length = 5
        self['creation-time'], algorithm = _struct.unpack(
            '>IB', data[offset: offset + length])
        offset += length
        self['public-key-algorithm'] = self._public_key_algorithms[algorithm]
        if self['public-key-algorithm'].startswith('rsa '):
            o, self['public-modulus'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
            o, self['public-exponent'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
        elif self['public-key-algorithm'].startswith('dsa '):
            o, self['prime'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
            o, self['group-order'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
            o, self['group-generator'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
            o, self['public-key'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
        elif self['public-key-algorithm'].startswith('elgamal '):
            o, self['prime'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
            o, self['group-generator'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
            o, self['public-key'] = self._parse_multiprecision_integer(
                data[offset:])
            offset += o
        else:
            raise NotImplementedError(
                'algorithm-specific key fields for {}'.format(
                    self['public-key-algorithm']))
        return offset

    def to_bytes(self):
        pass


def packets_from_bytes(data):
    offset = 0
    while offset < len(data):
        packet = PGPPacket()
        offset += packet.from_bytes(data=data[offset:])
        yield packet


def migrate(old_key, new_key):
    """Add the old key and sub-keys to the new key

    For example, to upgrade your master key, while preserving old
    signatures you'd made.  You will lose signature *on* your old key
    though, since sub-keys can't be signed (I don't think).
    """
    old_key_export = _get_stdout(
        ['gpg', '--export', old_key])
    old_key_packets = list(
        packets_from_bytes(data=old_key_export))
    if old_key_packets[0]['type'] != 'public-key packet':
        raise ValueError(
            '{} does not start with a public-key packet'.format(
                old_key))
    old_key_secret_export = _get_stdout(
        ['gpg', '--export-secret-keys', old_key])
    old_key_secret_packets = list(
        packets_from_bytes(data=old_key_secret_export))

    import pprint
    pprint.pprint(old_key_packets)
    pprint.pprint(old_key_secret_packets)


if __name__ == '__main__':
    import sys as _sys

    old_key, new_key = _sys.argv[1:3]
    migrate(old_key=old_key, new_key=new_key)
